<?php

class SyntaxHighlighter {
 
  private static function isChar($c) {
    return $c >= 'a' && $c <= 'z';
  }
  
  private static function isDigit($c) {
    return $c >= '0' && $c <= '9';
  }
 
  private static function inNumber($c) {
    return SyntaxHighlighter::isDigit($c) || $c == '.' ||
      // scientific notation
      $c == 'e' || $c == 'E' ||
      // floating-point numbers
      $c == 'f' || $c == 'F' || $c == 'd' || $c == 'D' ||
      // hexadecimal notation
      $c == 'x' || $c == 'X';
  }
 
  // all potential keywords [actually a subset: I don't need "all" of them]
  private static $keywords = array(
    "auto",
    "asm",
    "and",
    "abstract",
    "assert",
    "as",
    "array",
    
    "break",
    "bool",
    "boolean",
    "byte",
    "base",
    
    "case",
    "char",
    "const",
    "continue",
    "catch",
    "class",
    "catch",
    "checked",
    "clone",
    
    "default",
    "do",
    "double",
    "delete",
    "decimal",
    "delegate",
    "declare",
    
    "else",
    "enum",
    "extern",
    "export",
    "extends",
    "event",
    "explicit",
    "elseif",
    "enddeclare",
    "endfor",
    "endforeach",
    "endif",
    "endswitch",
    "endwhile",
    
    "float",
    "for",
    "false",
    "friend",
    "final",
    "finally",
    "fixed",
    "foreach",
    "final",
    "foreach",
    "function",
    
    "goto",
    "global",
    
    "if",
    "int",
    "inline",
    "implements",
    "import",
    "instanceof",
    "interface", 
    "implicit",
    "in",
    "internal",
    "is",
    "implements",
    "instanceof",
    
    "long",
    "lock",
    
    "mutable",
    
    "not",
    "namespace",
    "new",
    "native",
    "null",
    
    "or",
    "operator",
    "object",
    "out",
    "override",
    
    "private",
    "protected",
    "public",
    "package",
    "params",
    
    "register",
    "return",
    "readonly",
    "ref",
    
    "short",
    "signed",
    "sizeof",
    "static",
    "struct",
    "switch",
    "synchronized",
    "super",
    "struct",
    "sbyte",
    "sealed",
    "sizeof",
    "stackalloc",
    "string",
    
    "typedef",
    "true",
    "template",
    "this",
    "throw",
    "try",
    "typeid",
    "typename",
    "throws",
    "transient",
    "try",
    "typeof",
    
    "union",
    "unsigned",
    "using",
    "uint",
    "ulong",
    "unchecked",
    "unsafe",
    "ushort",
    "use",
    
    "void",
    "volatile",
    "virtual",
    "var",
    
    "while",
    
    "typeof",
    
    "xor",
    
    // PHP & R specials
    "__CLASS__",
    "__DIR__",
    "__FILE__",
    "__LINE__",
    "__FUNCTION__",
    "__METHOD__",
    "__NAMESPACE__",
    "FALSE",
    "TRUE",
    "NULL",
    "NA",
    "Inf",
    "NaN");

  // code highlighting targeting C-style languages (C, C++, java, C#, PHP): good enough for me.
  // Alternative: use a library like GeSHi [http://qbnz.com/highlighter/]
  public static function highlight($code) {
  
    // Method = finite state automate
    
    // still TODO: special coloration for objects, i.e. 'new BlaBla()'
    
    $c = 0; // current position in text
    
    $parsed = ""; // result after all substitutions
    
    $state = ParseState::UNDEFINED;
    
    $codeLength = strlen($code);
    while ($c < $codeLength) {
    
      // position to next non-empty char:
      while ($code[$c] == ' ') {
        $parsed .= "&nbsp;";
        $c++;
      }
      while ($code[$c] == '\n') {
        $parsed .= "<br/>\n";
        $c++;
      }
      while ($code[$c] == '\t') {
        // HACK: one tab == 4 empty spaces (to be refined ?)
        $parsed .= "&nbsp;&nbsp;&nbsp;&nbsp;";
        $c++;
      }
  
      switch ($state) {
        
        case ParseState::UNDEFINED:
        
          // start special modes ?
          if ($code[$c] == '"') {
            $state = ParseState::STRING_DBLQUOTE;
            $parsed .= "<span class='string_dblquote'>\"";
            $c++;
          }
          else if ($code[$c] == '\'') {
            $state = ParseState::STRING_DBLQUOTE;
            $parsed .= "<span class='string_sglquote'>'";
            $c++;
          }
          else if ($code[$c] == '#') {
            $state = ParseState::PREPROCESSING;
            $parsed .= "<span class='preprocessing'>#";
            $c++;
          }
          else if ($code[$c] == '/' && $c < $codeLength - 1 && $code[$c+1] == '*') {
            $state = ParseState::COMMENT_C;
            $parsed .= "<span class='comment_c'>/*";
            $c+=2;
          }
          else if ($code[$c] == '/' && $c < $codeLength - 1 && $code[$c+1] == '/') {
            $state = ParseState::COMMENT_CPP;
            $parsed .= "<span class='comment_cpp'>//";
            $c+=2;
          }
          else {
            // 'simple' code:
            $state = ParseState::CODE;
          }
          break;
        
        case ParseState::CODE:
          
          if (SyntaxHighlighter::isChar($code[$c])) {
            // variable, or function, or keyword
            $begin = $c++;
            while ($c < $codeLength && (SyntaxHighlighter::isChar($code[$c]) || SyntaxHighlighter::isDigit($code[$c]))) {
              $c++;
            }
            $end = $c;
            
            $word = substr($code, $begin, $end-$begin);
            if (in_array($word, SyntaxHighlighter::$keywords)) {
              $parsed .= "<span class='keyword'>".$word."</span>";
            }
            else if ($end < $codeLength && $code[$end] == '(') {
              $parsed .= "<span class='function'>".$word."</span>";
            }
            else {
              $parsed .= "<span class='variable'>".$word."</span>";
            }
          }
          
          else if (SyntaxHighlighter::isDigit($code[$c])) {
            // number
            $begin = $c++;
            while ($c < $codeLength && SyntaxHighlighter::inNumber($c)) {
              $c++;
            }
            $end = $c;
            
            $value = substr($code, $begin, $end-$begin);
            $parsed .= "<span class='number'>".$value."</span>";
          }
          
          else if ($code[$c] == ',' || $code[$c] == ';') {
            // delimiters
            $parsed .= "<span class='delimiter'>".$code[$c]."</span>";
            $c++;
          }
          
          else if ($code[$c] == '(' || $code[$c] == ')' || 
            $code[$c] == '{' || $code[$c] == '}' ||
            $code[$c] == '[' || $code[$c] == ']')
          {
            $parsed .= "<span class='inblock'>".$code[$c]."</span>";
            $c++;
          }
          
          // boolean or binary operators + special char '@' (for C#)
          else if ($code[$c] == '&' && $c < $codeLength && $code[$c+1] == '&') {
            $parsed .= "<span class='operator'>&&</span>";
            $c+=2;
          }
          else if ($code[$c] == '|' && $c < $codeLength && $code[$c+1] == '|') {
            $parsed .= "<span class='operator'>||</span>";
            $c+=2;
          }
          else if ($code[$c] == '@' || $code[$c] == '&' || $code[$c] == '|' || 
            $code[$c] == '^' || $code[$c] == '!' || $code[$c] == '=')
          {
            $parsed .= "<span class='operator'>".$code[$c]."</span>";
            $c++;
          }
          
          else {
            // any unrecognized thing !
            $parsed .= $code[$c];
            $c++;
          }
          
          break;
    
        case ParseState::COMMENT_C:
          if ($code[$c] == '*' && $c < $codeLength && $code[$c+1] == '/') {
            $state = ParseState::UNDEFINED;
            $c += 2;
            $parsed .= "*/";
          }
          else $c++;
          break;
              
        case ParseState::COMMENT_CPP:
          if ($code[$c] == '\n') {
            $state = ParseState::UNDEFINED;
            $parsed .= "<br/>\n";
          }
          $c++;
          break;
          
        case ParseState::PREPROCESSING:
          if ($code[$c] == '\n') {
            $state = ParseState::UNDEFINED;
            $parsed .= "<br/>\n";
          }
          $c++;
          break;
        
        case ParseState::STRING:
          // TODO: distinguish escape character.
          // TODO: distinguish regular expressions.
          $parsed .= $code[$c];
          $c++;
          break;
      }
    }
  
    // end of file reached: we could need to close some style
    if ($state != ParseState::UNDEFINED && $state != ParseState::CODE) {
      $parsed .= "</span>";
    }
  
    return $parsed;
  }
}

class ParseState {
  const UNDEFINED = 0;
  const CODE = 1;
  const COMMENT_C = 2;
  const COMMENT_CPP = 3;
  const PREPROCESSING = 4;
  const STRING_DBLQUOTE = 5;
  const STRING_SGLQUOTE = 6;
}

?>
